package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2013 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;
import java.util.Collections;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * This class enables sorting lists of Filament and FilamentChain objects. This
 * defines an injective mapping of lists of Filaments into lists of
 * FilamentChains and a well-defined mapping of lists of FilamentChains onto
 * lists of Filaments.
 */
public abstract class AbstractFilament implements Cloneable, Comparable<AbstractFilament>
{

    public abstract Point getCenter();

    public abstract boolean getKeep();

    public abstract long getLength();

    public abstract long getOrientation();

    public abstract long getWidth();

    public abstract void setKeep(boolean keep);

    public abstract void setWidth(long width);

    public abstract void invert();

    public abstract void remedyOldAngles();

    public abstract Element toXML(Document document, int chain_number, int filament_number);

    @Override
    public abstract Object clone();

    @Override
    public abstract boolean equals(Object obj);

    @Override
    public abstract int hashCode();

    @Override
    public int compareTo(AbstractFilament filament)
    {
        if (getWidth() > filament.getWidth())
        {
            return 1;
        }

        if (getWidth() < filament.getWidth())
        {
            return -1;
        }

        if (getWidth() == filament.getWidth())
        {
            if (getLength() > filament.getLength())
            {
                return 1;
            }

            if (getLength() < filament.getLength())
            {
                return -1;
            }
            if (getLength() == filament.getLength())
            {
                if (getCenter().y > filament.getCenter().y)
                {
                    return 1;
                }

                if (getCenter().y < filament.getCenter().y)
                {
                    return -1;
                }

                if (getCenter().y == filament.getCenter().y)
                {
                    if (getCenter().x > filament.getCenter().x)
                    {
                        return 1;
                    }

                    if (getCenter().x < filament.getCenter().x)
                    {
                        return -1;
                    }

                    if (getCenter().x == filament.getCenter().x)
                    {
                        if (getOrientation() > filament.getOrientation())
                        {
                            return 1;
                        }

                        if (getOrientation() < filament.getOrientation())
                        {
                            return -1;
                        }
                    }
                }
            }
        }
        return 0;
    }

    /**
     * This method takes a list as argument and has no return value. As the
     * elements of the list are changed (permuted), the argument itself is
     * changed by the method and thus the return is implicit.
     *
     * The list is first sorted and then reversed so as to begin with the
     * "largest" element.
     */
    public static void sortList(List<? extends AbstractFilament> list)
    {
        Collections.sort(list);
        Collections.reverse(list);
    }

}
