package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

public class DataParameters implements Cloneable
{

    private Map<Any, Integer> m_preprocessing_parameters, m_binarization_parameters,
                              m_elog_parameters, m_hough_parameters,
                              m_tracing_parameters, m_chaining_parameters,
                              m_orientation_parameters;
    private List<Map<Any, Integer>> m_maps;
    public static final int PRE = 1, ELOG = 2, HOUGH = 3, BIN = 4, TRACE = 5, CHAIN = 6, ORI = 7;
    public static final int[] ALL = {PRE, BIN, TRACE, ELOG, CHAIN, ORI};

    private List<Map<Any, Integer>> makeList()
    {
        List<Map<Any, Integer>> maps = new ArrayList<>();
        maps.add(m_preprocessing_parameters);
        maps.add(m_binarization_parameters);
        maps.add(m_elog_parameters);
        maps.add(m_hough_parameters);
        maps.add(m_tracing_parameters);
        maps.add(m_chaining_parameters);
        maps.add(m_orientation_parameters);
        return maps;
    }

    public DataParameters(Map<Any, Integer> pre_params, Map<Any, Integer> bin_params,
                          Map<Any, Integer> elog_params, Map<Any, Integer> hough_params,
                          Map<Any, Integer> trace_params, Map<Any, Integer> chain_params,
                          Map<Any, Integer> ori_params)
    {
        m_preprocessing_parameters = (pre_params   == null ? preprocessingDefaults() : pre_params);
        m_binarization_parameters  = (bin_params   == null ? binarizationDefaults()  : bin_params);
        m_elog_parameters          = (elog_params  == null ? elogDefaults()          : elog_params);
        m_hough_parameters         = (hough_params == null ? houghDefaults()         : hough_params);
        m_tracing_parameters       = (trace_params == null ? tracingDefaults()       : trace_params);
        m_chaining_parameters      = (chain_params == null ? chainingDefaults()      : chain_params);
        m_orientation_parameters   = (ori_params == null   ? orientationDefaults()   : ori_params);
        m_maps = makeList();
    }

    public DataParameters()
    {
        this(null, null, null, null, null, null, null);
    }

    public int getParameter(Any key)
    {
        Map<Any, Integer> map = getMap(key.type());
        if (map == null || !map.containsKey(key))
        {
            System.err.println("Key " + key + " was not found in map.");
            return -10000;
        }
        return map.get(key);
    }
    
    public HashMap<Any, Integer> getParameters(int type)
    {
        Map<Any, Integer> map = getMap(type);
        if (map == null)
        {
            return null;
        }
        return cloneHashMap((HashMap<Any, Integer>) map);  
    }
    
    public Set<Any> keySet(int type)
    {
        Map<Any, Integer> map = getMap(type);
        if (map == null)
        {
            return null;
        }
        return map.keySet(); 
    }
    
    public boolean containsKey(int type, Any key)
    {
        Map<Any, Integer> map = getMap(type);
        if (map == null)
        {
            return false;
        }
        return map.containsKey(key);
    }

    public void setParameter(Any key, int number)
    {
        Map<Any, Integer> map = getMap(key.type());
        if (map == null)
        {
            return;
        }
        map.put(key, number);  
    }
    
    public void setParameters(int type, HashMap<Any, Integer> params)
    {
        Map<Any, Integer> map = getMap(type);
        if (map == null)
        {
            return;
        }
        map = cloneHashMap(params);
    }

    public DataParameters completeFrom(DataParameters that)
    {
        for (int i = that.m_maps.size()-1; i >= 0; i--)
        {
            if (that.m_maps.get(i) == null)
            {
                continue;
            }

            if (m_maps.get(i) == null)
            {
                m_maps.set(i, that.m_maps.get(i));
                continue;
            }

            for (Any key : that.m_maps.get(i).keySet())
            {
                if (!m_maps.get(i).containsKey(key) && that.m_maps.get(i).containsKey(key))
                {
                    m_maps.get(i).put(key, that.m_maps.get(i).get(key));
                    System.out.println("Added " + key + " to parameters.");
                }
            }
        }
        return this;
    }

    public static String getName(int type)
    {
        switch (type)
        {
        case DataParameters.PRE:
            return "pre_params";
        case DataParameters.ELOG:
            return "elog_params";
        case DataParameters.HOUGH:
            return "hough_params";
        case DataParameters.BIN:
            return "bin_params";
        case DataParameters.TRACE:
            return "trc_params";
        case DataParameters.CHAIN:
            return "chain_params";
        case DataParameters.ORI:
            return "ori_params";
        default:
            System.err.println("DataParameters.getName(): Unknown parameter type " + type);
            return null;
        }   
    }

    public static DataParameters fromXML(Document document, DataParameters defaults)
    {
        HashMap<Any, Integer> pre_params = new HashMap<>();
        HashMap<Any, Integer> bin_params = new HashMap<>();
        HashMap<Any, Integer> trace_params = new HashMap<>();
        NamedNodeMap nodes;
        Element element;

        // Preprocessing parameters
        NodeList list = document.getElementsByTagName("pre_params");
        if (list.getLength() > 0)
        {
            element = (Element) list.item(0);

            nodes = element.getAttributes();
            for (int i = 0; i < nodes.getLength(); i++)
            {
                Attr a = (Attr) nodes.item(i);

                try
                {
                    pre_params.put(Pre.valueOf(a.getName()), Integer.parseInt(a.getValue()));
                }
                catch (NumberFormatException e)
                {
                    System.err.println(a.getName() + " has invalid value " + a.getValue());
                }
            }
        }

        // Binarization parameters
        list = document.getElementsByTagName("bin_params");
        if (list.getLength() > 0)
        {
            element = (Element) list.item(0);

            nodes = element.getAttributes();
            for (int i = 0; i < nodes.getLength(); i++)
            {
                Attr a = (Attr) nodes.item(i);

                try
                {
                    bin_params.put(Bin.valueOf(a.getName()), Integer.parseInt(a.getValue()));
                }
                catch (NumberFormatException e)
                {
                    System.err.println(a.getName() + " has invalid value " + a.getValue());
                }
            }
        }

        // Tracing parameters
        list = document.getElementsByTagName("trc_params");
        if (list.getLength() > 0)
        {
            element = (Element) list.item(0);

            nodes = element.getAttributes();
            for (int i = 0; i < nodes.getLength(); i++)
            {
                Attr a = (Attr) nodes.item(i);

                try
                {
                    trace_params.put(Trace.valueOf(a.getName()), Integer.parseInt(a.getValue()));
                }
                catch (NumberFormatException e)
                {
                    System.err.println(a.getName() + " has invalid value " + a.getValue());
                }
            }
        }
        return ((pre_params.isEmpty() && bin_params.isEmpty() && trace_params.isEmpty()) ? defaults :
                new DataParameters(pre_params, bin_params, null, null, trace_params,
                                   null, null)).completeFrom(defaults);
    }

    private Map<Any, Integer> getMap(int type)
    {
        switch (type)
        {
        case DataParameters.PRE:
            return m_preprocessing_parameters;
        case DataParameters.ELOG:
            return m_elog_parameters;
        case DataParameters.HOUGH:
            return m_hough_parameters;
        case DataParameters.BIN:
            return m_binarization_parameters;
        case DataParameters.TRACE:
            return m_tracing_parameters;
        case DataParameters.CHAIN:
            return m_chaining_parameters;
        case DataParameters.ORI:
            return m_orientation_parameters;
        default:
            System.err.println("DataParameters.getMap(): Unknown parameter type " + type);
            return null;
        }   
    }

    private static Map<Any, Integer> preprocessingDefaults()
    {
        Map<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Pre.minbright, 5);
        hash_map.put(Pre.contrast, 0);
        hash_map.put(Pre.black, -1);
        hash_map.put(Pre.white, -1);
        hash_map.put(Pre.rm_sng_px, 1);
        hash_map.put(Pre.do_laplace, 1);
        hash_map.put(Pre.lpfac, 20);
        hash_map.put(Pre.lpmask, 1);
        hash_map.put(Pre.do_gauss, 1);
        hash_map.put(Pre.fsigma, 10);
        hash_map.put(Pre.order, 0);
        hash_map.put(Pre.do_line_gauss, 1);
        hash_map.put(Pre.line_sigma, 50);
        hash_map.put(Pre.cross_corr, 0);
        hash_map.put(Pre.corr_mask_size, 3);
        hash_map.put(Pre.corr_zero, 35);
        hash_map.put(Pre.widths, 3);
        hash_map.put(Pre.preview, 1);
        return hash_map;
    }

    private static Map<Any, Integer> binarizationDefaults()
    {
        Map<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Bin.method, 1);
        hash_map.put(Bin.radius, 3);
        hash_map.put(Bin.sigma, 20);
        hash_map.put(Bin.minmean, 25);
        hash_map.put(Bin.area_significance, 100);
        hash_map.put(Bin.rod_significance, 50);
        hash_map.put(Bin.rm_sng_px, 1);
        hash_map.put(Bin.thicken, 0);
        return hash_map;
    }

    private static Map<Any, Integer> elogDefaults()
    {
        Map<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Elog.input, 0);
        hash_map.put(Elog.image, 0);
        hash_map.put(Elog.tolerance, 0);
        hash_map.put(Elog.zero_crossings, 0);
        hash_map.put(Elog.threshold, -1);
        hash_map.put(Elog.contrast, 100);
        hash_map.put(Elog.orientations, 18);
        hash_map.put(Elog.sigma_x, 100);
        hash_map.put(Elog.sigma_y, 7);
        return hash_map;
    }

    private static Map<Any, Integer> houghDefaults()
    {
        Map<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Hough.image, 2);
        hash_map.put(Hough.orientations, 179);
        hash_map.put(Hough.sigma, 100);
        hash_map.put(Hough.minmean, 50);
        return hash_map;
    }

    private static Map<Any, Integer> tracingDefaults()
    {
        Map<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Trace.minlen, 20);
        hash_map.put(Trace.minangle, 20);
        hash_map.put(Trace.tolerance, 5);
        return hash_map;
    }

    private static Map<Any, Integer> chainingDefaults()
    {
        HashMap<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Chain.dist_max, 10);
        hash_map.put(Chain.dist_max_gap, 2);
        hash_map.put(Chain.skew_max, 15);
        hash_map.put(Chain.gap_skew_max, 20);
        return hash_map;
    }

    private static Map<Any, Integer> orientationDefaults()
    {
        HashMap<Any, Integer> hash_map = new HashMap<>();
        hash_map.put(Ori.min_area, 50);
        hash_map.put(Ori.min_filaments, 10);
        return hash_map;
    }

    @Override
    public Object clone()
    {
        HashMap<Any, Integer> pre_dup   = cloneHashMap((HashMap<Any, Integer>) m_preprocessing_parameters);
        HashMap<Any, Integer> bin_dup   = cloneHashMap((HashMap<Any, Integer>) m_binarization_parameters);
        HashMap<Any, Integer> elog_dup  = cloneHashMap((HashMap<Any, Integer>) m_elog_parameters);
        HashMap<Any, Integer> hough_dup = cloneHashMap((HashMap<Any, Integer>) m_hough_parameters);
        HashMap<Any, Integer> trace_dup = cloneHashMap((HashMap<Any, Integer>) m_tracing_parameters);
        HashMap<Any, Integer> chain_dup = cloneHashMap((HashMap<Any, Integer>) m_chaining_parameters);
        HashMap<Any, Integer> ori_dup   = cloneHashMap((HashMap<Any, Integer>) m_orientation_parameters);
        DataParameters dup = new DataParameters(pre_dup, bin_dup, elog_dup, hough_dup, trace_dup,
                                                chain_dup, ori_dup);
        return dup;
    }

    private static HashMap<Any, Integer> cloneHashMap(HashMap<Any, Integer> hashmap)
    {
        HashMap<Any, Integer> dup = new HashMap<Any, Integer>();
        Set<Any> key_set = hashmap.keySet();
        for (Any key : key_set)
        {
            Integer entry_dup = hashmap.get(key).intValue();
            Any key_dup = key;
            dup.put(key_dup, entry_dup);
        }
        return dup;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((m_binarization_parameters  == null) ? 0 : m_binarization_parameters.hashCode());
        result = prime * result + ((m_elog_parameters          == null) ? 0 : m_elog_parameters.hashCode());
        result = prime * result + ((m_hough_parameters         == null) ? 0 : m_hough_parameters.hashCode());
        result = prime * result + ((m_preprocessing_parameters == null) ? 0 : m_preprocessing_parameters.hashCode());
        result = prime * result + ((m_tracing_parameters       == null) ? 0 : m_tracing_parameters.hashCode());
        result = prime * result + ((m_chaining_parameters      == null) ? 0 : m_chaining_parameters.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object object)
    {
        if (this == object)
        {
            return true;
        }

        if (object == null || this.getClass() != object.getClass())
        {
            return false;
        }

        DataParameters that = (DataParameters) object;

        if (m_binarization_parameters == null && that.m_binarization_parameters != null)
        {
            return false;
        }
        if (!m_binarization_parameters.equals(that.m_binarization_parameters))
        {
            return false;
        }

        if (m_elog_parameters == null && that.m_elog_parameters != null)
        {
            return false;
        }
        if (!m_elog_parameters.equals(that.m_elog_parameters))
        {
            return false;
        }

        if (m_hough_parameters == null && that.m_hough_parameters != null)
        {
            return false;
        }
        if (!m_hough_parameters.equals(that.m_hough_parameters))
        {
            return false;
        }

        if (m_preprocessing_parameters == null && that.m_preprocessing_parameters != null)
        {
            return false;
        }
        if (!m_preprocessing_parameters.equals(that.m_preprocessing_parameters))
        {
            return false;
        }

        if (m_tracing_parameters == null && that.m_tracing_parameters != null)
        {
            return false;
        }
        if (!m_tracing_parameters.equals(that.m_tracing_parameters))
        {
            return false;
        }

        if (m_chaining_parameters == null && that.m_chaining_parameters != null)
        {
            return false;
        }
        if (!m_chaining_parameters.equals(that.m_chaining_parameters))
        {
            return false;
        }

        return true;
    }

    public static interface Any{ public int type(); public String name(); }

    public enum Pre implements Any
    {
        minbright, contrast, black, white, rm_sng_px, do_laplace, lpfac,
        lpmask, do_gauss, fsigma, order, do_line_gauss, line_sigma,
        cross_corr, corr_mask_size, corr_zero, widths, preview;

        public int type(){ return PRE; }
    }

    public enum Bin implements Any
    {
        method, radius, sigma, minmean, area_significance,
        rod_significance, rm_sng_px, thicken;

        public int type(){ return BIN; }
    }

    public enum Elog implements Any
    {
        input, image, tolerance, zero_crossings, threshold, contrast,
        orientations, sigma_x, sigma_y;

        public int type(){ return ELOG; }
    }

    public enum Hough implements Any
    {
        image, orientations, sigma, minmean;

        public int type(){ return HOUGH; }
    }

    public enum Trace implements Any
    {
        minlen, minangle, tolerance;

        public int type(){ return TRACE; }
    }

    public enum Chain implements Any
    {
        dist_max, dist_max_gap, skew_max, gap_skew_max;

        public int type(){ return CHAIN; }
    }

    public enum Ori implements Any
    {
        min_area, min_filaments;

        public int type(){ return ORI; }
    }
}
