package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Color;
import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JOptionPane;

import ui.Config;
import ui.Dialogs;
import ui.FilamentSensor;
import ui.IOImageData;
import ui.IOProjectData;
import ui.MainWindow;

public abstract class DataProject
{
    protected static List<Files> m_files;
    protected static List<DataImage> m_processed_images;
    protected static boolean m_force_defaults;
    protected static DataParameters m_default_parameters;
    protected static Map<Integer, File> m_file_dictionary;

    public static boolean getForce()
    {
        return m_force_defaults;
    }

    public static void setForce(boolean force)
    {
        m_force_defaults = force;
    }

    public static DataParameters getDefaultParameters()
    {
        return (DataParameters) m_default_parameters.clone();
    }

    public static void setDefaultParameters(DataParameters params)
    {
        m_default_parameters = ((DataParameters) params.clone()).completeFrom(new DataParameters());
    }

    private static Color colorDialog(String message)
    {
        HashMap<String, Color> map = new HashMap<String, Color>();
        map.put("Orange", Color.orange);
        map.put("Dark Orange", new Color(255, 140, 0));
        map.put("Cyan", Color.cyan);
        map.put("Blue", Color.blue);

        Object[] options = map.keySet().toArray();
        String color_name = (String) JOptionPane.showInputDialog(MainWindow.frame(), message, "Choose Color",
                                    JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
        return map.get(color_name);
    }

    /**
     * This method traces filaments for all image files in m_files.
     */
    public static void batchProcessing(DataImage caller)
    {
        // Save dialogs for csv and filament images.
        File csv_path = Dialogs.save(true, "Folder to save csv.");
        Color color = colorDialog("Choose the color for filaments in the output images.");
        File image_path = Dialogs.save(true, "Folder to save images with filaments.");

        if (csv_path == null || image_path == null)
        {
            return;
        }

        Config.set("LastFolders", "Save", csv_path.getAbsolutePath());

        // Batch
        for (Files pair : m_files)
        {
            DataImage data = null;

            // Prefer data from the current session, default to data from XML files.
            for (DataImage data_changed : m_processed_images)
            {
                if (pair.image().getName().equals(data_changed.getImageName()))
                {
                    data = data_changed;
                }
            }
            if (data == null)
            {
                if (pair.data() == null)
                {
                    data = new DataImage(pair.image(), getDefaultParameters());
                }
                else
                {
                    IOImageData file = new IOImageData(pair.data());
                    data = file.read(pair.image().getParent(), getDefaultParameters(), getForce());
                }
            }

            data.updatePreprocessing();
            data.updateBinarization();
            data.scanFilaments();
            data.exportCSV(csv_path, false, false);
            data.saveImage(image_path, "filaments", DataImage.ORIG, true, color);

            if (data.getImageName().equals(caller.getImageName()))
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Identified " + data.getImageName() + " as caller.");
                }
                caller = data;
            }
            else
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Shutting down " + data.getImageName());
                }
                data.shutDown();
            }
        }
    }

    public static void batchElog(DataImage caller)
    {
        // Save dialogs for csv and filament images.
        File csv_path = Dialogs.save(true, "Folder to save histogram csv.");
        File image_path = Dialogs.save(true, "Folder to save eLoG images.");

        if (csv_path == null || image_path == null)
        {
            return;
        }

        Config.set("LastFolders", "Save", csv_path.getAbsolutePath());

        // Batch
        for (Files pair : m_files)
        {
            DataImage data = null;

            // Prefer data from the current session, default to data from XML files.
            for (DataImage data_changed : m_processed_images)
            {
                if (pair.image().getName().equals(data_changed.getImageName()))
                {
                    data = data_changed;
                }
            }
            if (data == null)
            {
                if (pair.data() == null)
                {
                    data = new DataImage(pair.image(), getDefaultParameters());
                }
                else
                {
                    IOImageData file = new IOImageData(pair.data());
                    data = file.read(pair.image().getParent(), getDefaultParameters(), getForce());
                }
            }

            String parameters = data.calculateELoGs();
            data.exportHistogram(csv_path, false, parameters, DataImage.ELOG_COLOR);
            data.saveImage(image_path, "eLoG", DataImage.ELOG_COLOR, false, null);

            if (data.getImageName().equals(caller.getImageName()))
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Identified " + data.getImageName() + " as caller.");
                }
                caller = data;
            }
            else
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Shutting down " + data.getImageName());
                }
                data.shutDown();
            }
        }
    }

    public static void batchHough(DataImage caller)
    {
        // Save dialogs for csv and filament images.
        File csv_path = Dialogs.save(true, "Folder to save histogram csv.");
        File image_path = Dialogs.save(true, "Folder to save Hough images.");

        if (csv_path == null || image_path == null)
        {
            return;
        }

        Config.set("LastFolders", "Save", csv_path.getAbsolutePath());

        // Batch
        for (Files pair : m_files)
        {
            DataImage data = null;

            // Prefer data from the current session, default to data from XML files.
            for (DataImage data_changed : m_processed_images)
            {
                if (pair.image().getName().equals(data_changed.getImageName()))
                {
                    data = data_changed;
                }
            }
            if (data == null)
            {
                if (pair.data() == null)
                {
                    data = new DataImage(pair.image(), getDefaultParameters());
                }
                else
                {
                    IOImageData file = new IOImageData(pair.data());
                    data = file.read(pair.image().getParent(), getDefaultParameters(), getForce());
                }
            }

            data.updatePreprocessing();
            data.updateBinarization();
            String parameters = data.calculateHough();
            data.exportHistogram(csv_path, false, parameters, DataImage.HOUGH);
            data.saveImage(image_path, "Hough", DataImage.HOUGH, false, null);

            if (data.getImageName().equals(caller.getImageName()))
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Identified " + data.getImageName() + " as caller.");
                }
                caller = data;
            }
            else
            {
                if (FilamentSensor.DEBUG)
                {
                    System.out.println("Shutting down " + data.getImageName());
                }
                data.shutDown();
            }
        }
    }

    public static void updateProject()
    {
        IOProjectData project = new IOProjectData(true);
        String image_path = project.getImagePath().getAbsolutePath();
        for (int n = 0; n < project.getLength(); n++)
        {
            File data = project.getDataFile(n);
            if (data != null)
            {
                IOImageData xml_file = new IOImageData(data);
                DataImage image_data = xml_file.read(image_path, DataProject.getDefaultParameters(), false);
                xml_file.writeImageDataFile(image_data);
                image_data.shutDown();
                System.out.println(data.getName() + " updated.");
            }
        }
        JOptionPane.showMessageDialog(MainWindow.frame(), "Upgrade complete. Returning to FilamentSensor.",
                                      "Project updated!", JOptionPane.INFORMATION_MESSAGE, null);
    }
    
    /**
     * This is a simple pairing class to enforce one-to-one correspondence of
     * image files with their respective XML data files.
     */
    public static class Files implements Comparable<Files>
    {
        private File image;
        private File data;

        public Files(File image, File data)
        {
            this.image = image;
            this.data = data;
        }

        public File image()
        {
            return image;
        }

        public File data()
        {
            return data;
        }

        public void setData(File data_file)
        {
            data = data_file;
        }

        @Override
        public int compareTo(Files that)
        {
            return image().compareTo(that.image());
        }
    }
}
