package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.util.Map;

import core.DataParameters.Any;
import core.DataParameters.Bin;
import core.DataParameters.Elog;
import core.DataParameters.Hough;
import core.DataParameters.Pre;
import ui.FilamentSensor;
import ui.ImageFactory;
import ui.ImageFileChooser;

public class ImageProcessing
{
    private int[][] m_int_image_initial,
                    m_int_image_preprocessed,
                    m_int_image_filter_difference,
                    m_int_image_elog, m_int_image_elog_alt;
    private boolean[][] m_binary_image,
                        m_binary_image_hough;
    
    private int[] m_elog_histogram,
                  m_hough_histogram;

    private int m_width, m_height, m_auto_black, m_auto_white, m_auto_elog_threshold;

    private BufferedImage m_buffered_image_original;
    private BufferedImage m_buffered_image_elog;

    private int m_preprocessing_parameters_hash,
                m_binarization_parameters_hash,
                m_elog_parameters_hash,
                m_hough_parameters_hash;

    private boolean m_update_elogs,
                    m_update_hough,
                    m_update_binarization;

    ImageProcessing(File img_file)
    {
        m_buffered_image_original = ImageFileChooser.readImage(img_file);

        if(FilamentSensor.DEBUG)
        {
            System.out.println("Image in original size was successfully read in by ImageProcessing!");
        }

        m_buffered_image_elog = null;
        m_update_elogs = false;
        m_update_hough = false;
        m_update_binarization = false;

        m_preprocessing_parameters_hash = 0;
        m_binarization_parameters_hash = 0;

        m_int_image_preprocessed = null;
        m_int_image_elog = null;
        m_int_image_elog_alt = null;
        m_int_image_filter_difference = null;
        m_binary_image_hough = null;
        m_binary_image = null;
        m_elog_histogram = null;
        m_hough_histogram = null;
        m_width = -1;
        m_height = -1;

        if (m_buffered_image_original == null)
        {
            return;
        }

        m_width = m_buffered_image_original.getWidth();
        m_height = m_buffered_image_original.getHeight();

        if (m_width <= 0 || m_height <= 0)
        {
            return;
        }

        m_int_image_preprocessed = new int[m_width][m_height];
        m_binary_image = new boolean[m_width][m_height];
        m_int_image_initial = makeIntImage();
        m_auto_elog_threshold = (new GlobalMethods(m_int_image_initial, true, GlobalMethods.LI)).threshold;
    }

    String getResolution()
    {
        return m_width + "x" + m_height;
    }

    int[][] getOriginalImage()
    {
        return m_int_image_initial;
    }

    CellData getCellData()
    {
        int[][] tmp = m_int_image_initial;
        tmp = gaussFilter(tmp, 0.1); // Strangely, this gives a 1.5 second speed up.
        tmp = gaussFilter(tmp, 2.0);
        GlobalMethods li = new GlobalMethods(tmp, true, GlobalMethods.LI);
        while (tooWhite(li.bin, 0.5) && li.threshold > 0)
        {
            new GlobalMethods(tmp, li.threshold, 255); // Rescales tmp!
            li = new GlobalMethods(tmp, true, GlobalMethods.LI);
        }
        return new CellData(li.bin);
    }

    boolean[][] getBinaryImage()
    {
        return m_binary_image;
    }
    
    int[] getBlackWhite()
    {
        return new int[]{ m_auto_black, m_auto_white };
    }

    int getElogThreshold()
    {
        return m_auto_elog_threshold;
    }

    String getHistogram(int type)
    {
        int[] histogram = null;
        switch (type)
        {
        case DataImage.ELOG_COLOR:
            histogram = m_elog_histogram;
            break;
        case DataImage.HOUGH:
            histogram = m_hough_histogram;
            break;
        default:
            return null;
        }
        String out = "";
        for (int i = 0; i < histogram.length; i++)
        {
            out += (i * 180 / (double)histogram.length) + "," + histogram[i] + "\n";
        }
        return out;
    }

    boolean updatePreprocessing(Map<Any, Integer> pre_params, boolean[][] cell_area, boolean changed)
    {
        if (FilamentSensor.DEBUG)
        {
            System.out.println("Update preprocessing.");
        }

        if (pre_params.hashCode() != m_preprocessing_parameters_hash)
        {
            m_int_image_preprocessed = new int[m_width][m_height];
            m_int_image_filter_difference = new int[m_width][m_height];
            int black = 0, white = 255;
            
            if (pre_params.get(Pre.contrast) == 0)
            {
                black = m_auto_black;
                white = m_auto_white;
            }
            else
            {
                black = pre_params.get(Pre.black);
                white = pre_params.get(Pre.white);
            }

            m_int_image_preprocessed = enhanceContrast(m_int_image_initial,
                                                       black, white);
//            applyBrightThresh(pre_params.get("minbright"));
            if (pre_params.get(Pre.rm_sng_px) != 0)
            {
                cleanPixelsPre(m_int_image_preprocessed);
            }

            if (pre_params.get(Pre.order) == 0)
            {
                if (pre_params.get(Pre.do_gauss) != 0)
                {
                    m_int_image_preprocessed = gaussFilter(m_int_image_preprocessed,
                                                           pre_params.get(Pre.fsigma) / 10.0);
                }
                if (pre_params.get(Pre.do_laplace) != 0)
                {
                    m_int_image_preprocessed = laplaceFilter(m_int_image_preprocessed,
                                                             pre_params.get(Pre.lpmask),
                                                             pre_params.get(Pre.lpfac) / 10.0);
                }
            }
            else
            {
                if (pre_params.get(Pre.do_laplace) != 0)
                {
                    m_int_image_preprocessed = laplaceFilter(m_int_image_preprocessed,
                                                             pre_params.get(Pre.lpmask),
                                                             pre_params.get(Pre.lpfac) / 10.0);
                }
                if (pre_params.get(Pre.do_gauss) != 0)
                {
                    m_int_image_preprocessed = gaussFilter(m_int_image_preprocessed,
                                                           pre_params.get(Pre.fsigma) / 10.0);
                }
            }

            if (pre_params.get(Pre.cross_corr) != 0)
            {
                double[][][] masks = makeMasks(1, pre_params.get(Pre.corr_mask_size));
                m_int_image_preprocessed = crossCorrelationFilter(m_int_image_preprocessed, masks,
                                                                  pre_params.get(Pre.corr_zero) / 100.0);
            }

            if (pre_params.get(Pre.do_line_gauss) != 0)
            {
                m_int_image_preprocessed = lineGaussFilter(m_int_image_preprocessed,
                                                           pre_params.get(Pre.line_sigma) / 10.0,
                                                           cell_area);
            }

            for (int x = 0; x < m_width; x++)
            {
                for (int y = 0; y < m_height; y++)
                {
                    m_int_image_filter_difference[x][y] = m_int_image_preprocessed[x][y] -
                                                          m_int_image_initial[x][y];
                }
            }

            // invalidate binarization
            m_binarization_parameters_hash = 0;
            changed = true;
        }

        m_preprocessing_parameters_hash = pre_params.hashCode();
        m_update_elogs = true;
        m_update_binarization = true;
        return changed;
    }

    boolean updateBinarization(Map<Any, Integer> bin_params, boolean elog_null)
    {
        if (FilamentSensor.DEBUG)
        {
            System.out.println("Update binarization.");
        }

        if (bin_params.hashCode() == m_binarization_parameters_hash && !m_update_binarization)
        {
            return false;
        }

        switch (bin_params.get(Bin.method))
        {
        case 0:
            m_binary_image = areaBinarization(m_int_image_preprocessed,
                                              bin_params.get(Bin.minmean),
                                              bin_params.get(Bin.sigma) / 10.0,
                                              bin_params.get(Bin.area_significance) / 10.0);
            break;
        case 1:
            m_binary_image = rodBinarization(m_int_image_preprocessed,
                                             bin_params.get(Bin.minmean),
                                             bin_params.get(Bin.sigma) / 10.0,
                                             bin_params.get(Bin.rod_significance) / 10.0);
            break;
        default:
            return false;
        }

        if (bin_params.get(Bin.thicken) != 0)
        {
            thickenLines(m_binary_image);
        }
        else
        {
            cleanPixelsBin(m_binary_image);
        }

        m_binarization_parameters_hash = bin_params.hashCode();
        m_update_binarization = false;
        m_update_hough = true;
        return true;
    }

    boolean calculateELoGs(Map<Any, Integer> elog_params)
    {
        if ( (elog_params.hashCode() == m_elog_parameters_hash && !m_update_elogs) || m_int_image_initial == null)
        {
            return false;
        }

        m_int_image_elog = new int[m_width][m_height];
        m_int_image_elog_alt = new int[m_width][m_height];
        m_elog_histogram = new int[elog_params.get(Elog.orientations)];

        //laplacianImage();
        double[][][] gaussians = gaussians(elog_params.get(Elog.orientations),
                                           elog_params.get(Elog.sigma_x)/10.0,
                                           elog_params.get(Elog.sigma_y)/10.0);

        double[][][] elog_masks = new double[gaussians.length][][];

        // Invert order to correct angle definitions.
        for (int ori = 0; ori < gaussians.length; ori++)
        {
            elog_masks[ori] = convolve(gaussians[gaussians.length - 1 - ori],
                                       new double[][]{{0,-1,0},
                                                      {-1,4,-1},
                                                      {0,-1,0}});
        }

        int threshold = elog_params.get(Elog.threshold);
        threshold = (threshold < 0 ? m_auto_elog_threshold : threshold);
        eLoGmain(elog_params.get(Elog.tolerance), elog_params.get(Elog.contrast), elog_masks, threshold);

        m_buffered_image_elog = ImageFactory.makeColorImage(m_int_image_elog,
                                                            elog_params.get(Elog.orientations));
        m_elog_parameters_hash = elog_params.hashCode();
        m_update_elogs = false;

        return true;
    }

    boolean calculateHough(Map<Any, Integer> hough_params)
    {
        if (hough_params.hashCode() == m_hough_parameters_hash && !m_update_hough)
        {
            return false;
        }

        int[][] hough_image = null;

        switch (hough_params.get(Hough.image))
        {
        case 2:
            hough_image = houghTransform(m_binary_image, hough_params.get(Hough.orientations));
            break;
        default:
            return false;
        }

        //new PopUpDebug(makeGreyImage(rescaleImage(hough_image), false));
        m_binary_image_hough = findMaxima(hough_image);
        makeHoughHistogram(hough_image, m_binary_image_hough);
        
        m_hough_parameters_hash = hough_params.hashCode();
        m_update_hough = false;
        return true;
    }

    int scoreImage(boolean[][] mask, int sigma)
    {
        final double ori_threshold = 1.1 / (Math.sqrt(2 * Math.PI) * sigma);
        final int bri_threshold = 200;
        if (mask == null)
        {
            mask = getCellData().getMask();
        }

        final int width = m_int_image_initial.length,
                  height = m_int_image_initial[0].length,
                  size = 3*sigma;
        final double[][] gauss = Calc.gaussianMask(size, sigma);
        final Point[][] lines = latticeInit(size);

        double gauss_total = 0;
        for (int a = 0; a < 2 * size + 1; a++)
        {
            for (int b = 0; b < 2 * size + 1; b++)
            {
                gauss_total += gauss[a][b];
            }
        }

        int area = 0;
        double bright_count = 0,
               oriented_count = 0;
        for (int x = size; x < width-size; x++)
        {
            for (int y = size; y < height-size; y++)
            {
                if (mask[x][y])
                {
                    // Gaussian weighted mean
                    double gauss_mean = 0;
                    for (int dx = -size; dx < size+1; dx++)
                    {
                        for (int dy = -size; dy < size+1; dy++)
                        {
                            gauss_mean += m_int_image_initial[x+dx][y+dy] * gauss[size+dx][size+dy];
                        }
                    }
                    gauss_mean /= gauss_total;
                    
                    // Maximal Gaussian weighted rod mean.
                    double max_ori_mean = 0;
                    for (int ori=0; ori<4*size; ori++)
                    {
                        double sum =  m_int_image_initial[x][y] * gauss[size][size];
                        //double gauss_sum = gauss[size][size];
                        double mean = 0.0;

                        for (int dist=1; dist< size+1; dist++)
                        {
                            int dx = lines[ori][dist].x;
                            int dy = lines[ori][dist].y;

                            sum += (m_int_image_initial[x+dx][y+dy] * gauss[size+dx][size+dy] +
                                    m_int_image_initial[x-dx][y-dy] * gauss[size-dx][size-dy]);
                            //gauss_sum += (gauss[size+dx][size+dy] + gauss[size-dx][size-dy]);
                        }
                        mean = sum / gauss_total;

                        if(mean > max_ori_mean)
                        {
                            max_ori_mean = mean;
                        }
                    }

                    // Calculate measures
                    area++;
                    if (max_ori_mean / gauss_mean > ori_threshold)
                    {
                        oriented_count++;
                    }
                    if (gauss_mean > bri_threshold)
                    {
                        bright_count++;
                    }
                }
            }
        }
        oriented_count /= (float)area;
        bright_count /= (float)area;
        return (int)Math.round(0.001 * area - 7500 * bright_count + 1500 * oriented_count);
    }

    private static boolean tooWhite (boolean[][] binary_image, double ratio)
    {
        final int width = binary_image.length,
                  height = binary_image[0].length,
                  white_threshold = (int)(width * height * ratio),
                  black_threshold = width * height - white_threshold;
        int white_pixels = 0;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    white_pixels++;
                }
            }
            if (white_pixels > white_threshold)
            {
                return true;
            }
            if (x * height - white_pixels > black_threshold)
            {
                return false;
            }
        }
        return false;
    }

    private static int[][] houghTransform(boolean[][] bool_image, int orientations)
    {
        int width = bool_image.length;
        int half_width = width / 2;
        int height = bool_image[0].length;
        int half_height = height / 2;
        int half_diagonal = (int)Math.ceil(Math.sqrt(half_width * half_width +
                                                     half_height * half_height));
        double step = Math.PI / (double) orientations;
        int[][] hough_image = new int[2 * half_diagonal + 1][orientations + 20];

        for (int x = 1 - half_width; x < half_width; x++)
        {
            for (int y = 1 - half_height; y < half_height; y++)
            {
                if (!bool_image[x + half_width - 1][y + half_height - 1])
                {
                    for (int i = - 10; i < orientations + 10; i++)
                    {
                        // Orientation inversion using +sin.
                        int dist = (int) Math.round(Math.sin(i * step) * x + Math.cos(i * step) * y);
                        hough_image[half_diagonal + dist][i+10]++;
                    }
                }
            }
        }
        return hough_image;
    }

    private static boolean[][] findMaxima(int[][] image)
    {
        int width = image.length;
        int height = image[0].length;
        int[][] tmp_image = new int[width][height];
        boolean[][] maxima = new boolean[width][height];
        int radius = 1;
        for (int x = radius; x < width - radius; x++)
        {
            for (int y = radius; y < height - radius; y++)
            {
                for (int i = -radius; i <= radius; i++)
                {
                    for (int j = -radius; j <= radius; j++)
                    {
                        tmp_image[x][y] += image[x+i][y+j];
                    }
                }
            }
        }
        int range = (2*radius+1);
        for (int x = range; x < width - range; x++)
        {
            for (int y = range; y < height - range; y++)
            {
                maxima[x][y] = true;
                for (int i = -range; i <= range; i+= range)
                {
                    for (int j = -range; j <= range; j+= range)
                    {
                        if (i == 0 && j == 0)
                        {
                            continue;
                        }
                        maxima[x][y] = maxima[x][y] && (tmp_image[x][y] >= tmp_image[x+i][y+j]) && tmp_image[x][y] != 0;
                    }
                }
            }
        }
        
        int[][] max_image = new int[width][height];
        int counter = 1;

        for (int t = 0; counter > 0; t ++)
        {
            counter = 0;
            for (int x = range; x < width - range; x++)
            {
                for (int y = range; y < height - range; y++)
                {
                    maxima[x][y] = true;
                    for (int i = -range; i <= range; i+= range)
                    {
                        for (int j = -range; j <= range; j+= range)
                        {
                            if (i == 0 && j == 0)
                            {
                                continue;
                            }
                            maxima[x][y] = maxima[x][y] && tmp_image[x][y] != 0 &&
                                           (tmp_image[x][y] >= tmp_image[x+i][y+j] + t * range * range);
                        }
                    }
                    if (maxima[x][y])
                    {
                        max_image[x][y]++;
                        counter++;
                    }
                }
            }
        }
        maxima = (new GlobalMethods(max_image, true, GlobalMethods.OTSU)).bin;

        return maxima;
    }

    private void makeHoughHistogram(int[][] image, boolean[][] mask)
    {
        int width = image.length;
        int height = image[0].length;
        m_hough_histogram = new int[height - 20];

        for (int x = 0; x < width; x++)
        {
            for (int y = 10; y < height - 10; y++)
            {
                if (mask[x][y])
                {
                    m_hough_histogram[y - 10] += image[x][y];
                }
            }
        }
    }

    private static void cleanPixelsPre(int[][] image)
    {
        if (image == null || image.length <= 0 || image[0].length <= 0)
        {
            return;
        }

        int width = image.length;
        int height = image[0].length;

        for (int i = 1; i < width - 1; i++)
        {
            for (int j = 1; j < height - 1; j++)
            {
                if (image[i][j] != 0 && image[i - 1][j] == 0 && image[i][j - 1] == 0 &&
                    image[i + 1][j] == 0 && image[i][j + 1] == 0)
                {
                    image[i][j] = 0;
                }
            }
        }
    }

    private int[][] makeIntImage()
    {
        Raster raster = m_buffered_image_original.getData();

        int max = 0;
        int min = Integer.MAX_VALUE;
        int[][] int_image = new int[m_width][m_height];
        for (int i = 0; i < m_width; i++)
        {
            for (int j = 0; j < m_height; j++)
            {
                int pixel = int_image[i][j] = raster.getSample(i, j, 0);

                if (pixel > max)
                {
                    max = pixel;
                }
                if (pixel < min)
                {
                    min = pixel;
                }
            }
        }
        if (FilamentSensor.DEBUG)
        {
            System.out.println("min = " + min + ". max = " + max);
        }
        
        GlobalMethods contrast = new GlobalMethods(int_image, min, max);
        m_auto_black = contrast.black;
        m_auto_white = contrast.white;

        return int_image;
    }

    private static int[][] enhanceContrast(int[][] image, int black, int white)
    {
        double bw_scale = 255 / ((double) (white - black));
        int width = image.length;
        int height = image[0].length;
        int[][] out = new int[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                int pixel = image[i][j];
                pixel = (int) Math.round(bw_scale * (pixel - black));
                pixel = (pixel < 0 ? 0 : pixel);
                pixel = (pixel > 255 ? 255 : pixel);

                out[i][j] = pixel;
            }
        }
        return out;
    }

    private void eLoGmain(int tolerance, int contrast, double[][][] masks, int threshold)
    {
        int orientations = masks.length;

        if (orientations == 0)
        {
            return;
        }

        if (FilamentSensor.DEBUG)
        {
            System.out.println("Processing eLoGs...");
        }

        double total_max = Double.MIN_VALUE;
        double total_min = Double.MAX_VALUE;
        double[][] tmp = new double[m_width][m_height];
        for (int i = 0; i < m_width; i++)
        {
            for (int j = 0; j < m_height; j++)
            {
                if (m_int_image_initial[i][j] > threshold)
                {
                    ELoGData elog = new ELoGData(i, j, masks, false);
                    m_int_image_elog[i][j] = elog.max_ori;
                    tmp[i][j] = elog.max;

                    if(elog.max > total_max)
                    {
                        total_max = elog.max;
                    }
                    if(elog.max < total_min)
                    {
                        total_min = elog.max;
                    }
                }
                else
                {
                    m_int_image_elog[i][j] = -1;
                }
            }

            if (FilamentSensor.DEBUG)
            {
                if (i % (m_width / 10) == 0)
                {
                    System.out.println( ( 100 * i / m_width ) + "%");
                }
            }
        }

        for (int i = 0; i < m_width; i++)
        {
            for (int j = 0; j < m_height; j++)
            {
                if (m_int_image_initial[i][j] > threshold)
                {
                    m_int_image_elog_alt[i][j] = Math.min(255, (int)( contrast * 25.5 *
                                                                      (tmp[i][j] - total_min) /
                                                                      (total_max - total_min) ));
                }
                else
                {
                    m_int_image_elog_alt[i][j] = 0;
                }
            }
        }
        
        boolean[][] elog_image_otsu = (new GlobalMethods(m_int_image_elog_alt, true, 1, GlobalMethods.OTSU)).bin;

        for (int i = 0; i < m_width; i++)
        {
            for (int j = 0; j < m_height; j++)
            {
                if (!elog_image_otsu[i][j])
                {
                    m_int_image_elog[i][j] = -1;
                }
                else if (m_int_image_elog[i][j] != -1)
                {
                    m_elog_histogram[m_int_image_elog[i][j]]++;
                }
            }
        }

        if (FilamentSensor.DEBUG)
        {
            System.out.println("100% done.");
        }
    }

    /**
     * Applies a gaussian filter to m_image_preprocessed. The size of the
     * convolution mask is chosen such that a maximally bright boundary
     * contributes at most one unit of brightness.
     */
    private static int[][] gaussFilter(int[][] image, double sigma)
    {
        if (sigma == 0.0)
        {
            return image;
        }
        // Find range such that the boundary is not too bright.
        int range = 2;
        while (8. * 255. * range * Math.exp(-range * range/(2.0 * sigma * sigma)) /
               (2.0 * Math.PI * sigma * sigma) > 1)
        {
            range++;
        }

        double[][] mask = Calc.gaussianMask(range, sigma);
        final int width = image.length,
                  height = image[0].length;
        int out[][] = new int[width][height];

        // Left and right boundary
        for (int i = 0; i < width; i++)
        {
            if (i == range)
            {
                i = width - range;
            }
            for (int j = 0; j < height; j++)
            {
                double sum = 0.0;
                double bias = 0.0;

                for (int a = -range; a <= range; a++)
                {
                    for (int b = -range; b <= range; b++)
                    {
                        if (i+a >= 0 && i+a < width && j+b >=0 && j+b < height)
                        {
                            sum += mask[a + range][b + range] * image[i + a][j + b];
                            bias += mask[a + range][b + range];
                        }
                    }
                }
                out[i][j] = (int) (sum / bias);
            }
        }
        // Upper and lower boundary
        for (int j = 0; j < height; j++)
        {
            if (j == range)
            {
                j = height - range;
            }
            for (int i = 0; i < width; i++)
            {
                double sum = 0.0;
                double bias = 0.0;

                for (int a = -range; a <= range; a++)
                {
                    for (int b = -range; b <= range; b++)
                    {
                        if (i+a >= 0 && i+a < width && j+b >=0 && j+b < height)
                        {
                            sum += mask[a + range][b + range] * image[i + a][j + b];
                            bias += mask[a + range][b + range];
                        }
                    }
                }

                out[i][j] = (int) (sum / bias);
            }
        }
        // Interior
        for (int i = range; i < width - range; i++)
        {
            for (int j = range; j < height - range; j++)
            {
                double sum = 0.0;
                double bias = 0.0;

                for (int a = -range; a <= range; a++)
                {
                    for (int b = -range; b <= range; b++)
                    {
                        sum += mask[a + range][b + range] * image[i + a][j + b];
                        bias += mask[a + range][b + range];
                    }
                }

                out[i][j] = (int) (sum / bias);
            }
        }
        return out;
    }

    /**
     * Applies a laplace filter to m_image_preprocessed. The convolution mask is chosen from a
     * predefined list and multiplied by lpfac.
     */
    private static int[][] laplaceFilter(int[][] image, int type, double factor)
    {
        if (image == null || image.length <= 0 || image[0].length <= 0)
        {
            return null;
        }

        int[][] mask;

        switch (type)
        {
        case 0:
            mask = new int[][]
            {
            {  0, -1,  0 },
            { -1,  4, -1 },
            {  0, -1,  0 } };
            break;
        case 1:
            mask = new int[][]
            {
            { -1, -1, -1 },
            { -1,  8, -1 },
            { -1, -1, -1 } };
            break;
        default:
            if (FilamentSensor.DEBUG)
            {
                System.out.println("Unknown laplace mask-type: " + type);
            }
            return null;
        }

        int width = image.length;
        int height = image[0].length;
        int tmp[][] = new int[width][height];

        for (int i = 0; i < width; i++)
        {
            tmp[i][0] = image[i][0];
            tmp[i][height-1] = image[i][height-1];
        }
        for (int i = 0; i < height; i++)
        {
            tmp[0][i] = image[0][i];
            tmp[width-1][i] = image[width-1][i];
        }

        for (int i = 1; i < width - 1; i++)
        {
            for (int j = 1; j < height - 1; j++)
            {
                int sum = 0;

                for (int a = -1; a <= 1; a++)
                {
                    for (int b = -1; b <= 1; b++)
                    {
                        sum += mask[a + 1][b + 1] * image[i + a][j + b];
                    }
                }
                tmp[i][j] = Math.min(Math.max(image[i][j] + (int) (factor * sum), 0), 255);
            }
        }

        return tmp;
    }

    private static int[][] lineGaussFilter(int[][] image, double sigma, boolean[][] cell_area)
    {
        final int range = (int)Math.ceil(3 * sigma);
        final double[][] mask = Calc.gaussianMask(range, sigma);
        final Point[][] lines = latticeInit(range);
        final int width = image.length,
                  height = image[0].length;
        int tmp[][] = new int[width][height];

        for (int x = range; x < width-range; x++)
        {
            for (int y = range; y < height-range; y++)
            {
                double max_ori_mean = 0;

                if(cell_area[x][y])
                {
                    for (int ori=0; ori<4*range; ori++)
                    {
                        double sum =  image[x][y] * mask[range][range];
                        double gauss_sum = mask[range][range];
                        double mean = 0.0;

                        for (int dist=1; dist<=range; dist++)
                        {
                            int dx = lines[ori][dist].x;
                            int dy = lines[ori][dist].y;

                            sum += (image[x+dx][y+dy] * mask[range+dx][range+dy] +
                                    image[x-dx][y-dy] * mask[range-dx][range-dy]);
                            gauss_sum += (mask[range+dx][range+dy] + mask[range-dx][range-dy]);
                        }
                        mean = sum / gauss_sum;

                        if(mean > max_ori_mean)
                        {
                            max_ori_mean = mean;
                        }
                    }
                    tmp[x][y] = (int)max_ori_mean;
                }
                else
                {
                    tmp[x][y] = image[x][y];
                }
            }
        }
        return tmp;
    }

    private static boolean[][] areaBinarization(int[][] image, int minimum_mean,
                                                double sigma, double significance)
    {
        significance *= significance;
        final int mean_range = (int) (Math.ceil(2 * sigma)),
                  gauss_range = (int) (Math.ceil(3 * sigma)),
                  step = (2 * mean_range + 1) * (2 * mean_range + 1) * minimum_mean;

        final double[][] gauss = Calc.gaussianMask(gauss_range, sigma);

        final int width = image.length,
                  height = image[0].length;
        int[][] squares = new int[width][height];
        boolean[][] out = new boolean[width][height];

        double[] weights = new double[256];
        for (int i = 0; i < weights.length; i++)
        {
            if (i < 30)
            {
                weights[i] = (double)i * i;
            }
            else
            {
                weights[i] = (4. * (-i + 36 * Math.sqrt(i) - 100)) / 9.;
                weights[i] *= weights[i];
            }
        }

        for (int i = 0; i < gauss_range; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[i][j] = true;
                out[width-1-i][j] = true;
            }
        }
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < gauss_range; j++)
            {
                out[i][j] = true;
                out[i][height-1-j] = true;
            }
        }

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                squares[i][j] = image[i][j] * image[i][j];
            }
        }

        for (int x = gauss_range; x < width-gauss_range; x++)
        {
            int bright = 0;
            for (int y = gauss_range; y < height-gauss_range; y++)
            {
                // Neighborhood mean thresholding
                if (y == gauss_range)
                {
                    bright = 0;
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        for(int b = -mean_range; b <= mean_range; b++)
                        {
                            bright += image[x+a][y+b];
                        }
                    }
                }
                else
                {
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        bright += (image[x+a][y+mean_range] -
                                   image[x+a][y-mean_range-1]);
                    }
                }

                if(bright < step)
                {
                    out[x][y] = true;
                    continue;
                }
                
                // Gaussian adaptive means
                double gauss_mean = 0.0;
                double sum_gauss = 0.0;
                
                for (int a = -gauss_range; a <= gauss_range; a++)
                {
                    for (int b = -gauss_range; b <= gauss_range; b++)
                    {
                        int i = x + a;
                        int j = y + b;
                
                        gauss_mean += (gauss[a + gauss_range][b + gauss_range] * image[i][j]);
                        sum_gauss += gauss[a + gauss_range][b + gauss_range];
                    }
                }
                
                gauss_mean /= sum_gauss;
                if (image[x][y] <= gauss_mean)
                {
                    out[x][y] = true;
                    continue;
                }

                // Statistical area filter
                double gauss_squares = 0;
            
                for (int a = -gauss_range; a <= gauss_range; a++)
                {
                    for (int b = -gauss_range; b <= gauss_range; b++)
                    {
                        int i = x + a;
                        int j = y + b;
            
                        gauss_squares += (squares[i][j] * gauss[a + gauss_range][b + gauss_range]);
                    }
                }
                out[x][y] = (10000.0 * gauss_squares / sum_gauss <=
                             10000.0 * gauss_mean * gauss_mean +
                             weights[(int) gauss_mean] * significance);
            }
        }
        return out;
    }


    private static boolean[][] rodBinarization(int[][] image, int minimum_mean,
                                                       double sigma, double significance)
    {
        final int mean_range = (int) (Math.ceil(2 * sigma)),
                  gauss_range = (int) (Math.ceil(3 * sigma)),
                  step = (2 * mean_range + 1) * (2 * mean_range + 1) * minimum_mean,
                  width = image.length,
                  height = image[0].length;

        final double[][] gauss = Calc.gaussianMask(gauss_range, sigma);
        final Point[][] lines = latticeInit(gauss_range);

        double[] weights = new double[256];
        for (int i = 0; i < weights.length; i++)
        {
            if (i < 30)
            {
                weights[i] = 1. / (double)i;
            }
            else
            {
                weights[i] = 9. / (4. * (-i + 36 * Math.sqrt(i) - 100));
            }
        }

        boolean[][] out = new boolean[width][height];

        for (int i = 0; i < gauss_range; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[i][j] = true;
                out[width-1-i][j] = true;
            }
        }
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < gauss_range; j++)
            {
                out[i][j] = true;
                out[i][height-1-j] = true;
            }
        }

        for (int x = gauss_range; x < width-gauss_range; x++)
        {
            int bright = 0;
            for (int y = gauss_range; y < height-gauss_range; y++)
            {
                // Neighborhood mean thresholding
                if (y == gauss_range)
                {
                    bright = 0;
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        for(int b = -mean_range; b <= mean_range; b++)
                        {
                            bright += image[x+a][y+b];
                        }
                    }
                }
                else
                {
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        bright += (image[x+a][y+mean_range] -
                                   image[x+a][y-mean_range-1]);
                    }
                }

                if(bright < step)
                {
                    out[x][y] = true;
                    continue;
                }

                // Gaussian adaptive means
                double sum = 0.0;
                double sum_gauss = 0.0;

                for (int a = -gauss_range; a <= gauss_range; a++)
                {
                    for (int b = -gauss_range; b <= gauss_range; b++)
                    {
                        int i = x + a;
                        int j = y + b;

                        sum += (gauss[a + gauss_range][b + gauss_range] * image[i][j]);
                        sum_gauss += gauss[a + gauss_range][b + gauss_range];
                    }
                }

                sum /= sum_gauss;
                if (image[x][y] <= sum)
                {
                    out[x][y] = true;
                    continue;
                }

                // Statistical rod filter
                double mean = 0;
                double standard_deviation = 0;

                for (int ori=0; ori<4*gauss_range; ori++)
                {
                    double rod_mean = image[x][y];

                    for (int dist=1; dist<=gauss_range; dist++)
                    {
                        int dx = lines[ori][dist].x,
                            dy = lines[ori][dist].y;

                        rod_mean += (image[x+dx][y+dy] + image[x-dx][y-dy]);
                    }

                    rod_mean /= (double)(2 * gauss_range + 1);
                    standard_deviation += rod_mean * rod_mean;
                    mean += rod_mean;
                }
                mean /= (double) (4 * gauss_range);
                standard_deviation = Math.sqrt(standard_deviation / (double) (4 * gauss_range) -
                                               mean * mean);
                
                out[x][y] = (100.0 * standard_deviation * weights[(int) mean] < significance);
            }
        }
        return out;
    }

    /**
     * Remove single, standalone pixels that are probably noise.
     */
    private static void cleanPixelsBin(boolean[][] image)
    {
        if (image == null)
        {
            return;
        }

        int width = image.length;
        int height = image[0].length;

        for (int i = 1; i < width - 1; i++)
        {
            for (int j = 1; j < height - 1; j++)
            {
                if (image[i][j] && !image[i - 1][j] && !image[i][j - 1]
                        && !image[i + 1][j] && !image[i][j + 1])
                {
                    image[i][j] = false;
                }
            }
        }
    }

    private static double[][][] gaussians(int orientations, double sigma_x, double sigma_y)
    {
        int range = (int)Math.ceil(3*Math.max(sigma_x, sigma_y));
        double phi_step = 180.0 / orientations;

        double[] phi = new double[orientations];
        double[][][] gaussians = new double[orientations][][];

        if (FilamentSensor.DEBUG)
        {
            System.out.println("Processing eLoGs...");
        }

        for (int ori = 0; ori < orientations; ori++)
        {
            phi[ori] = Math.toRadians(ori * phi_step);
            gaussians[ori] = Calc.gaussianMask(range, phi[ori], sigma_x, sigma_y);
        }

        return gaussians;
    }

    private static double[][] convolve(double[][] image, double[][] mask)
    {
        if (image.length <= 0 || image[0].length <= 0 || mask.length <= 0 || mask[0].length <= 0 ||
            image.length < mask.length || image[0].length < mask[0].length ||
            mask.length % 2 == 0 || mask[0].length % 2 == 0)
        {
            return null;
        }

        int width = image.length;
        int height = image[0].length;
        int range_horizontal = ( mask.length - 1 ) / 2;
        int range_vertical = ( mask[0].length - 1 ) / 2;

        double out[][] = new double[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < range_vertical; j++)
            {
                out[i][j] = image[i][j];
                out[i][height-j-1] = image[i][height-j-1];
            }
        }
        for (int j = 0; j < height; j++)
        {
            for (int i = 0; i < range_horizontal; i++)
            {
                out[i][j] = image[i][j];
                out[height-i-1][j] = image[height-i-1][j];
            }
        }
        for (int i = range_horizontal; i < width - range_horizontal; i++)
        {
            for (int j = range_vertical; j < height - range_vertical; j++)
            {
                for (int a = -range_horizontal; a <= range_horizontal; a++)
                {
                    for (int b = -range_vertical; b <= range_vertical; b++)
                    {
                        out[i][j] += mask[a + range_horizontal][b + range_vertical] * image[i + a][j + b];
                    }
                }
            }
        }
        return out;
    }

    private static int[][] crossCorrelationFilter(int[][] image, double[][][] masks, double zero_crossing)
    {
        if (image == null || masks == null ||image.length <= 0 || image[0].length <= 0 ||
            masks.length <= 0 || masks[0].length <= 0 || masks[0][0].length <= 0 ||
            image.length < masks[0].length || image[0].length < masks[0][0].length ||
            masks[0].length % 2 == 0 || masks[0][0].length % 2 == 0)
        {
            return image;
        }

        final int width = image.length,
                  height = image[0].length,
                  n_masks = masks.length,
                  range_horizontal = ( masks[0].length - 1 ) / 2,
                  range_vertical = ( masks[0][0].length - 1 ) / 2;
        final double area = (2 * range_horizontal + 1) * (2 * range_vertical + 1);

        // Precalculate prefactors from masks to save time
        double[] factor1 = new double[n_masks], factor2 = new double[n_masks];
        for (int m = 0; m < n_masks; m++)
        {
            double sum = 0.0, square_sum = 0.0;
            for (int a = 0; a <= 2 * range_horizontal; a++)
            {
                for (int b = 0; b <= 2 * range_vertical; b++)
                {
                    sum += masks[m][a][b];
                    square_sum += masks[m][a][b] * masks[m][a][b];
                }
            }
            factor1[m] = area / Math.sqrt(area * square_sum - sum * sum);
            factor2[m] = sum / Math.sqrt(area * square_sum - sum * sum);
        }

        // Precalculate matrix of squares to save time
        double[][] squares = new double[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                squares[i][j] = image[i][j] * image[i][j];
            }
        }

        int[][] out = new int[width][height];

        // Edges of image remain unchanged
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < range_vertical; j++)
            {
                out[i][j] = image[i][j];
                out[i][height-j-1] = image[i][height-j-1];
            }
        }
        for (int j = 0; j < height; j++)
        {
            for (int i = 0; i < range_horizontal; i++)
            {
                out[i][j] = image[i][j];
                out[width-i-1][j] = image[width-i-1][j];
            }
        }

        // Calculate the maximum cross correlation for each point.
        double total_min = Double.MAX_VALUE;
        double total_max = Double.MIN_VALUE;
        double first_sum = 0.0;
        double first_square_sum = 0.0;
        for (int a = 0; a <= 2 * range_horizontal; a++)
        {
            for (int b = 0; b <= 2 * range_vertical; b++)
            {
                first_sum += image[a][b];
                first_square_sum += squares[a][b];
            }
        }

        for (int i = range_horizontal; i < width - range_horizontal; i++)
        {
            if (i > range_horizontal)
            {
                for (int b = 0; b <= 2 * range_vertical; b++)
                {
                    first_sum += (image[i+range_horizontal][b] -
                                  image[i-range_horizontal-1][b]);
                    first_square_sum += (squares[i+range_horizontal][b] -
                                         squares[i-range_horizontal-1][b]);
                }
            }
            double window_sum = first_sum;
            double window_square_sum = first_square_sum;

            for (int j = range_vertical; j < height - range_vertical; j++)
            {
                if (j > range_vertical)
                {
                    for (int a = - range_horizontal; a <= range_horizontal; a++)
                    {
                        window_sum += (image[i+a][j+range_vertical] -
                                       image[i+a][j-range_vertical-1]);
                        window_square_sum += (squares[i+a][j+range_vertical] -
                                              squares[i+a][j-range_vertical-1]);
                    }
                }

                double max = Double.MIN_VALUE;
                if (window_square_sum == 0.0)
                {
                    out[i][j] = 0;
                    continue;
                }

                double factor = 1 / Math.sqrt(area * window_square_sum - (double)window_sum * window_sum);
                for (int m = 0; m < n_masks; m++)
                {
                    double cross_sum = 0.0;
                    for (int a = - range_horizontal; a <=  range_horizontal; a++)
                    {
                        for (int b = - range_vertical; b <= range_vertical; b++)
                        {
                            cross_sum += image[i+a][j+b] *
                                         masks[m][range_horizontal + a][range_vertical + b];
                        }
                    }
                    double correlation = factor * (factor1[m] * cross_sum - factor2[m] * window_sum);
                    if (correlation > max)
                    {
                        max = correlation;
                    }
                }
                if (max < total_min)
                {
                    total_min = max;
                }
                if (max > total_max)
                {
                    total_max = max;
                }

                out[i][j] = (int)(Const.M * max);
            }
        }

        total_min += zero_crossing * (total_max + total_min);
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[i][j] = (int)(255 * (out[i][j] / Const.MF - total_min) / (total_max - total_min));
                if (out[i][j] < 0)
                {
                    out[i][j] = 0;
                }
            }
        }
        return out;
    }

    private static double[][][] makeMasks(int widths, int size)
    {
        if (size <= 0 || widths <= 0)
        {
            return null;
        }

        int range = size + widths;
        double[][][] temp = new double[4 * size * widths][2 * range + 1][2 * range + 1];
        double sigma = 0.5;
        double[][] gauss = Calc.gaussianMask(widths, sigma);

        for (int w = 0; w < widths; w++)
        {
            for (int dir = 0; dir < size; dir++)
            {
                double slope = (double)dir / (double)size;

                for (int x = - range; x <= range; x++)
                {
                    temp[dir + 4 * w * size][range + x][range + (int)Math.round(slope * x)] = 1;
                }
                temp[dir + 4 * w * size] = convolve(convolve(temp[dir + 4 * w * size], Const.CIRCLE[w+1]), gauss);
            }

            for (int dir=size; dir<3*size; dir++)
            {
                double epols = ( 2 * (double)size - dir ) / size;
                for (int y = - range; y <= range; y++)
                {
                    temp[dir + 4 * w * size][range + (int)Math.round(epols * y)][range + y] = 1;
                }
                temp[dir + 4 * w * size] = convolve(convolve(temp[dir + 4 * w * size], Const.CIRCLE[w+1]), gauss);
            }

            for (int dir=3*size; dir<4*size; dir++)
            {
                double slope = (dir - 4 * (double)size) / size;
                for (int x = - range; x <= range; x++)
                {
                    temp[dir + 4 * w * size][range + x][range + (int)Math.round(slope * x)] = 1;
                }
                temp[dir + 4 * w * size] = convolve(convolve(temp[dir + 4 * w * size], Const.CIRCLE[w+1]), gauss);
            }
        }
        double[][][] out = new double[4 * size * widths][2 * size + 1][2 * size + 1];
        for (int n = 0; n < 4 * size * widths; n++)
        {
            for (int x = 0; x <= 2 * size; x++)
            {
                for (int y = 0; y <= 2 * size; y++)
                {
                    out[n][x][y] = temp[n][x+widths][y+widths];
                }
            }
        }
        return out;
    }

    private static Point[][] latticeInit(int radius)
    {
        Point lines[][] = new Point[4*radius][radius+1];

        if (FilamentSensor.DEBUG)
        {
            System.out.println("Creating " + 4*radius + " lines.");
        }

        for (int direction = 0; direction < radius; direction++)
        {
            double slope = (double)direction / (double)radius;

            for(int x=0; x<=radius; x++)
            {
                lines[direction][x] = new Point(x, (int)Math.round(slope * x));
            }
        }

        for (int direction=radius; direction<3*radius; direction++)
        {
            double epols = ( 2 * (double)radius - direction ) / radius;
            for(int y=0; y<=radius; y++)
            {
                lines[direction][y] = new Point((int)Math.round(epols * y), y);
            }
        }

        for (int direction=3*radius; direction<4*radius; direction++)
        {
            double slope = (direction - 4 * (double)radius) / radius;
            for(int x=0; x<=radius; x++)
            {
                lines[direction][x] = new Point(x, (int)Math.round(slope * x));
            }
        }

        return lines;
    }



    private static void thickenLines(boolean[][] binary_image)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0)
        {
            return;
        }

        int width = binary_image.length;
        int height = binary_image[0].length;
        boolean[][] tmp = new boolean[width][height];

        for (int x = 1; x < width-1; x++)
        {
            for (int y = 1; y < height-1; y++)
            {
                if (!binary_image[x][y])
                {
                    tmp[x][y-1] = tmp[x-1][y] = tmp[x][y] = tmp[x+1][y] = tmp[x][y+1] = true;
                }
            }
        }

        for (int x = 1; x < width-1; x++)
        {
            for (int y = 1; y < height-1; y++)
            {
                if (tmp[x][y])
                {
                    binary_image[x][y] = false;
                }
            }
        }

        for (int x = 1; x < width-1; x++)
        {
            for (int y = 1; y < height-1; y++)
            {
                if (!tmp[x][y])
                {
                    binary_image[x][y-1] = binary_image[x-1][y] = binary_image[x][y]
                            = binary_image[x+1][y] = binary_image[x][y+1] = true;
                }
            }
        }
    }

    BufferedImage getImage(int type)
    {
        switch (type)
        {
        case DataImage.ORIG:
            return m_buffered_image_original;
        case DataImage.PRE:
            return ImageFactory.makeGreyImage(m_int_image_preprocessed, false);
        case DataImage.ELOG_COLOR:
            return m_buffered_image_elog;
        case DataImage.ELOG_GREY:
            return ImageFactory.makeGreyImage(m_int_image_elog_alt, false);
        case DataImage.DIFF:
            return ImageFactory.makeGreyImage(m_int_image_filter_difference, true);
        case DataImage.BIN:
            return ImageFactory.makeBinaryImage(m_binary_image, true);
        case DataImage.HOUGH:
            return ImageFactory.makeBinaryImage(m_binary_image_hough, false);
        default:
            return null;
        }
    }

    /**
     * Simple wrapper class for return values of eLoG.
     */
    private class ELoGData
    {
        private int max_ori = -1;
        private double min = Double.MAX_VALUE, max = 0;

        private ELoGData(int x, int y, double[][][] masks, boolean calc_min)
        {
            int orientations = masks.length;

            if (orientations == 0)
            {
                return;
            }
            
            int range = (masks[0].length - 1) / 2;

            double response[] = new double[orientations];

            for (int ori = 0; ori < orientations; ori++)
            {
                response[ori] = 0.0;

                for (int dx = -range; dx <= range; dx++)
                {
                    for (int dy = -range; dy <= range; dy++)
                    {
                        int a = x + dx;
                        int b = y + dy;

                        if (a >= 0 && a < m_width && b >= 0 && b < m_height)
                        {
                            response[ori] += m_int_image_initial[a][b] * masks[ori][range+dx][range+dy];
                        }
                    }
                }

                if (response[ori] > max)
                {
                    max = response[ori];
                    max_ori = ori;
                }
            }

            if(calc_min)
            {
                for(int z=0; z<=orientations/3; z++)
                {
                    int test_ori = (max_ori + orientations / 3 + z) % orientations;
                    if(response[test_ori] < min)
                    {
                        min = response[test_ori];
                    }
                }
            }
        }
    }

    private static class GlobalMethods
    {
        private boolean[][] bin;
        private int[][] trunkated;
        private int[] brightness_histogram;
        private int width, height, threshold, black, white;
        private static final int OTSU = 0, LI = 1;

        /**
         * This is the constructor to obtain the Otsu binarized image.
         */
        private GlobalMethods(int[][] int_image, boolean make_images, int lower_bound, int type)
        {
            if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
            {
                threshold = - 1;
                return;
            }

            width = int_image.length;
            height = int_image[0].length;

            brightness_histogram = makeHistogram(int_image, 0, 255);

            int number_of_pixels = width * height;
            if (lower_bound > 0)
            {
                number_of_pixels = 0;
                for (int i = lower_bound; i < 256; i++)
                {
                    number_of_pixels += brightness_histogram[i];
                }
                brightness_histogram[0] = number_of_pixels;
                for (int i = 1; i < lower_bound; i++)
                {
                    brightness_histogram[i] = 0;
                }
                number_of_pixels *= 2;
            }
            
            threshold = calcThreshold(number_of_pixels, 0, 255, type);

            if (make_images)
            {
                makeImages(int_image);
            }
        }
        
        private GlobalMethods(int[][] int_image, boolean make_images, int type)
        {
            this(int_image, make_images, 0, type);
        }

        /**
         * This constructor is for contrast enhancement.
         * MAY CHANGE INPUT int_image!
         */
        private GlobalMethods(int[][] int_image, int min, int max)
        {
            if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
            {
                threshold = - 1;
                return;
            }

            width = int_image.length;
            height = int_image[0].length;

            brightness_histogram = makeHistogram(int_image, min, max);

            // Get Otsu threshold as upper bound for black.
            int black_threshold = calcThreshold(width * height, 0, 255, OTSU);
            int new_number = 0;

            for (int i = black_threshold; i < 256; i++)
            {
                new_number += brightness_histogram[i];
            }

            // Get Otsu threshold for the part of the histogram above
            // black_threshold as a lower bound for white.
            int white_threshold = black_threshold;
            int n_over = new_number;
            while(new_number / n_over < 100 && white_threshold < 254) // This could be adjustable.
            {
                int tmp = white_threshold;
                white_threshold = calcThreshold(n_over, white_threshold, 255, OTSU);
                if (tmp == white_threshold)
                {
                    break;
                }

                n_over = 0;
                for (int i = white_threshold; i < 256; i++)
                {
                    n_over += brightness_histogram[i];
                }
            }

            calcBlackWhite(new_number);
            black = (black > black_threshold ? black_threshold :
                        (black < black_threshold / 2 ? black_threshold / 2 : black));
            white = (white < white_threshold ? white_threshold : white);
            white = (white < black + 128 ? black + 128 : white);
            white = (white > 255 ? 255 : white);

            if(FilamentSensor.DEBUG)
            {
                System.out.println("After adjustment: black = " + black + ". white = " + white + ".");
            }
        }

        /**
         * MAY CHANGE INPUT int_image!
         */
        private int[] makeHistogram(int[][] int_image, int min, int max)
        {
            int[] histogram = new int[256];
            double scale = 255 / ((double) (max - min));

            for (int i = 0; i < width; i++)
            {
                for (int j = 0; j < height; j++)
                {
                    if(min != 0 || max != 255)
                    {
                        int_image[i][j] = (int_image[i][j] < min ? 0 :
                                          (int_image[i][j] > max ? 255 :
                                           (int) Math.round(scale * (int_image[i][j] - min))));
                    }
                    histogram[int_image[i][j]]++;
                }
            }
            return histogram;
        }

        private int calcThreshold(int number_of_pixels, int min, int max, int type)
        {
            long sum_below = 0, sum_above = 0;
            double goodness_max = 0.0;

            long n_below = 0;
            long n_above = number_of_pixels;
            int thresh = 0;

            for (int i = min; i < max + 1; i++)
            {
                sum_above += (i + 1) * brightness_histogram[i];
            }

            switch (type)
            {
            case OTSU:
                if (FilamentSensor.DEBUG)
                {
                    System.out.print("Calculating Otsu threshold...");
                }
                for (int t = min; t < max; t++)
                {
                    n_below += brightness_histogram[t];
                    n_above -= brightness_histogram[t];

                    sum_below += (t + 1) * brightness_histogram[t];  // including t
                    sum_above -= (t + 1) * brightness_histogram[t];  // excluding t

                    double goodness = ((n_below * n_above) <= 0 ? 0 :
                                          (double) (n_below * sum_above - n_above * sum_below) *
                                          (double) (n_below * sum_above - n_above * sum_below) /
                                          (double) (n_below * n_above));

                    if (goodness > goodness_max)
                    {
                        goodness_max = goodness;
                        thresh = t;
                    }
                }
                break;
            case LI:
                if (FilamentSensor.DEBUG)
                {
                    System.out.print("Calculating Li threshold...");
                }
                double[] mean_below = new double[max - min], mean_above = new double[max - min];
                for (int t = min; t < max; t++)
                {
                    n_below += brightness_histogram[t];
                    n_above -= brightness_histogram[t];

                    sum_below += (t + 1) * brightness_histogram[t];  // including t
                    sum_above -= (t + 1) * brightness_histogram[t];  // excluding t

                    mean_below[t - min] = (n_below == 0 ? 0 : sum_below / (double) n_below);
                    mean_above[t - min] = (n_above == 0 ? mean_above[t - 1 - min] : sum_above / (double) n_above);
                }
                
                int t_old = -1, t_new = 1;
                while (t_old != t_new)
                {
                    t_old = t_new;
                    t_new = (Math.log(mean_below[t_old]) <= 0 ?
                             (int) Math.round(mean_above[t_old] / Math.log(mean_above[t_old])) :
                             (int) Math.round((mean_below[t_old] - mean_above[t_old]) /
                                              (Math.log(mean_below[t_old]) - Math.log(mean_above[t_old]))));
                }
                thresh = t_new - 1;
                break;
            default:
                thresh = 0;
            }

            if (FilamentSensor.DEBUG)
            {
                System.out.println("Threshold found: " + thresh);
            }

            return thresh;
        }

        private void makeImages(int[][] int_image)
        {
            bin = new boolean[width][height];
            trunkated = new int[width][height];

            for (int i = 0; i < width; i++)
            {
                for (int j = 0; j < height; j++)
                {
                    bin[i][j] = int_image[i][j] > threshold;
                    trunkated[i][j] = (int_image[i][j] > threshold ? int_image[i][j] : 0);
                }
            }

            // clean single pixels
            for (int i = 1; i < width - 1; i++)
            {
                for (int j = 1; j < height - 1; j++)
                {
                    if (bin[i][j] && !bin[i - 1][j] &&
                       !bin[i][j - 1] && !bin[i + 1][j] &&
                       !bin[i][j + 1])
                    {
                        bin[i][j] = false;
                        trunkated[i][j] = 0;
                    }
                }
            }
        }

        private void calcBlackWhite(int bright_pixels)
        {
            black = 0;
            white = 255;
            int step = bright_pixels / Const.CONTRAST_FACTOR;
            step = (step > 10 ? 10 : step);

            for (int i = 1; i < 255; i++)
            {
                if(FilamentSensor.VERBOSE)
                {
                    if ((i - 1) % 10 == 0)
                    {
                        System.out.print("Histogram " + i + "-" + (i+9) + ": ");
                    }
                    System.out.print(brightness_histogram[i] + ", ");
                    if (i % 10 == 0 || i == 254)
                    {
                        System.out.println("");
                    }
                }

                // Compensate for bright disturbances.
                if (brightness_histogram[i] > brightness_histogram[i + 1] + step && brightness_histogram[i + 1] != 0)
                {
                    white = i + 1;
                }

                // Compensate for bright background.
                if (brightness_histogram[254 - i] > brightness_histogram[255 - i] + 400 * step && brightness_histogram[255 - i] != 0)
                {
                    black = 254 - i;
                }
            }

            if(FilamentSensor.DEBUG)
            {
                System.out.println("Initial try: black = " + black + ". white = " + white + ". step = " + step);
            }
        }
    }

}
